import os
from datetime import date
from unittest.mock import patch

import matplotlib
import platform
import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
from covidcast import plotting

SHAPEFILE_PATHS = {
    "county": "../../covidcast/shapefiles/county/cb_2019_us_county_5m.shp",
    "state": "../../covidcast/shapefiles/state/cb_2019_us_state_5m.shp",
    "msa": "../../covidcast/shapefiles/msa/cb_2019_us_cbsa_5m.shp",
    "hrr": "../../covidcast/shapefiles/hrr/geo_export_ad86cff5-e5ed-432e-9ec2-2ce8732099ee.shp"}

CURRENT_PATH = os.path.dirname(os.path.realpath(__file__))

NON_GEOMETRY_COLS = ["geo_value", "time_value", "direction", "issue", "lag", "value", "stderr",
                     "sample_size", "geo_type", "data_source", "signal", "state_fips"]


@pytest.mark.skipif(platform.system() != "Linux", reason="Linux specific plot rendering expected.")
@patch("covidcast.plotting._signal_metadata")
def test_plot_choropleth(mock_metadata):
    mock_metadata.side_effect = [
        {"mean_value": 0.5330011, "stdev_value": 0.4683431},
        {"mean_value": 0.5330011, "stdev_value": 0.4683431},
        {"mean_value": 0.5304083, "stdev_value": 0.235302},
        {"mean_value": 0.5705364, "stdev_value": 0.4348706},
    ]
    matplotlib.use("agg")
    # load expected choropleth as an array
    expected = np.load(os.path.join(CURRENT_PATH, "../reference_data/expected_plot_arrays.npz"))

    # test county plots
    test_county = pd.read_csv(
        os.path.join(CURRENT_PATH, "../reference_data/test_input_county_signal.csv"), dtype=str)
    test_county["time_value"] = test_county.time_value.astype("datetime64[D]")
    test_county["value"] = test_county.value.astype("float")

    fig1 = plotting.plot_choropleth(test_county, time_value=date(2020, 8, 4))
    data1 = np.frombuffer(fig1.canvas.tostring_rgb(), dtype=np.uint8)  # get np array representation
    # give margin of +-2 for floating point errors and weird variations (1 isn't consistent)
    assert np.allclose(data1, expected["expected_1"], atol=2, rtol=0)

    fig2 = plotting.plot_choropleth(test_county, cmap="viridis", figsize=(5, 5), edgecolor="0.8")
    data2 = np.frombuffer(fig2.canvas.tostring_rgb(), dtype=np.uint8)
    assert np.allclose(data2, expected["expected_2"], atol=2, rtol=0)

    # test state
    test_state = pd.read_csv(
        os.path.join(CURRENT_PATH, "../reference_data/test_input_state_signal.csv"), dtype=str)
    test_state["time_value"] = test_state.time_value.astype("datetime64[D]")
    test_state["value"] = test_state.value.astype("float")
    fig3 = plotting.plot_choropleth(test_state)
    data3 = np.frombuffer(fig3.canvas.tostring_rgb(), dtype=np.uint8)
    assert np.allclose(data3, expected["expected_3"], atol=2, rtol=0)

    # test MSA
    test_msa = pd.read_csv(
        os.path.join(CURRENT_PATH, "../reference_data/test_input_msa_signal.csv"), dtype=str)
    test_msa["time_value"] = test_msa.time_value.astype("datetime64[D]")
    test_msa["value"] = test_msa.value.astype("float")
    fig4 = plotting.plot_choropleth(test_msa)
    data4 = np.frombuffer(fig4.canvas.tostring_rgb(), dtype=np.uint8)
    assert np.allclose(data4, expected["expected_4"], atol=2, rtol=0)


def test_get_geo_df():
    test_input = pd.DataFrame({"geo_value": ["24510", "31169", "37000"],
                               "value": [1.5, 2.5, 3],
                               "geo_type": ["county", "county", "county"],
                               "signal": ["a", "a", "a"],
                               "data_source": ["b", "b", "b"]})
    # test counties
    output1 = plotting.get_geo_df(test_input)
    expected1 = gpd.read_file(
        os.path.join(CURRENT_PATH, "../reference_data/expected_get_geo_df_right.gpkg"),
        dtype={"geo_value": str})
    pd.testing.assert_frame_equal(expected1, output1)

    # test counties w/ left join
    output2 = plotting.get_geo_df(test_input, join_type="left")
    expected2 = gpd.read_file(
        os.path.join(CURRENT_PATH, "../reference_data/expected_get_geo_df_left.gpkg"),
        dtype={"geo_value": str})
    pd.testing.assert_frame_equal(expected2, output2)

    # test states
    test_input["geo_type"] = "state"
    test_input["geo_value"] = ["24510", "31169", "ca"]
    output3 = plotting.get_geo_df(test_input)
    expected3 = gpd.read_file(
        os.path.join(CURRENT_PATH, "../reference_data/expected_get_geo_df_state.gpkg"),
        dtype={"geo_value": str})
    pd.testing.assert_frame_equal(expected3, output3)

    # test MSAs
    test_input["geo_type"] = "msa"
    test_input["geo_value"] = ["10420", "10580", "ca"]
    output4 = plotting.get_geo_df(test_input)
    expected4 = gpd.read_file(
        os.path.join(CURRENT_PATH, "../reference_data/expected_get_geo_df_msa.gpkg"),
        dtype={"geo_value": str})
    pd.testing.assert_frame_equal(expected4, output4)

    # test HRRs
    # test MSAs
    test_input["geo_type"] = "hrr"
    test_input["geo_value"] = ["10420", "102", "96"]
    output5 = plotting.get_geo_df(test_input)
    expected5 = gpd.read_file(
        os.path.join(CURRENT_PATH, "../reference_data/expected_get_geo_df_hrr.gpkg"),
        dtype={"geo_value": str})
    pd.testing.assert_frame_equal(expected5, output5)

    # test with sample signal
    test_signal = pd.read_csv(
        os.path.join(CURRENT_PATH, "../reference_data/test_input_county_signal.csv"),
        dtype={"geo_value": str}, parse_dates=["time_value", "issue"]
    )
    expected_geo_signal = gpd.read_file(
        os.path.join(CURRENT_PATH, "../reference_data/expected_get_geo_df_right_2.gpkg"),
        dtype={"geo_value": str})
    # geopandas reads file types slightly differently than pandas so need t`o recast
    expected_geo_signal["time_value"] = expected_geo_signal.time_value.astype("datetime64[ns]")
    expected_geo_signal["issue"] = expected_geo_signal.issue.astype("datetime64[ns]")
    expected_geo_signal["direction"] = np.nan
    output5 = plotting.get_geo_df(test_signal)
    pd.testing.assert_frame_equal(expected_geo_signal, output5)

    # test a non county or state geo_type
    with pytest.raises(ValueError):
        plotting.get_geo_df(pd.DataFrame(
            {"geo_value": ["a"], "geo_type": ["b"], "signal": ["c"], "data_source": ["d"]}))

    # test_duplicate_values
    with pytest.raises(ValueError):
        plotting.get_geo_df(pd.DataFrame({"geo_value": ["24510", "24510"]}))


def test__join_state_geo_df():
    test_input = pd.DataFrame({"state_code": ["ca", "al", "ak"],
                               "value": [1.5, 2.5, 3]})
    geo_info = gpd.read_file(os.path.join(CURRENT_PATH, SHAPEFILE_PATHS["state"]))
    # test right join
    output1 = plotting._join_state_geo_df(test_input, "state_code", geo_info)
    assert type(output1) is gpd.GeoDataFrame
    expected1 = gpd.read_file(
        os.path.join(CURRENT_PATH, "../reference_data/expected__join_state_geo_df_right.gpkg"),
        dtype={"geo_value": str})
    pd.testing.assert_frame_equal(expected1, output1)

    # test left join
    output2 = plotting._join_state_geo_df(test_input, "state_code", geo_info, "left")
    expected2 = gpd.read_file(
        os.path.join(CURRENT_PATH, "../reference_data/expected__join_state_geo_df_left.gpkg"),
        dtype={"geo_value": str})
    pd.testing.assert_frame_equal(expected2, output2)


def test__join_county_geo_df():
    test_input = pd.DataFrame({"county_code": ["24510", "31169", "37000"],
                               "test_value": [1.5, 2.5, 3],
                               "test_value2": [21.5, 32.5, 34]})
    geo_info = gpd.read_file(os.path.join(CURRENT_PATH, SHAPEFILE_PATHS["county"]))
    # test right join
    output1 = plotting._join_county_geo_df(test_input, "county_code", geo_info)
    assert type(output1) is gpd.GeoDataFrame
    expected1 = gpd.read_file(
        os.path.join(CURRENT_PATH, "../reference_data/expected__join_county_geo_df_right.gpkg"),
        dtype={"geo_value": str})
    pd.testing.assert_frame_equal(expected1, output1)

    # test left join
    output2 = plotting._join_county_geo_df(test_input, "county_code", geo_info, "left")
    expected2 = gpd.read_file(
        os.path.join(CURRENT_PATH, "../reference_data/expected__join_county_geo_df_left.gpkg"),
        dtype={"geo_value": str})
    pd.testing.assert_frame_equal(expected2, output2)


def test__join_msa_geo_df():
    test_input = pd.DataFrame({"msa": ["10180", "10420", "10580"],
                               "test_value": [1.5, 2.5, 3],
                               "test_value2": [21.5, 32.5, 34]})
    geo_info = gpd.read_file(os.path.join(CURRENT_PATH, SHAPEFILE_PATHS["msa"]))
    # test right join
    output1 = plotting._join_msa_geo_df(test_input, "msa", geo_info)
    assert type(output1) is gpd.GeoDataFrame
    # check that state parsing is working as intended
    assert all(output1[output1.msa == "35620"].state_fips == "36")

    expected1 = gpd.read_file(
        os.path.join(CURRENT_PATH, "../reference_data/expected__join_msa_geo_df_right.gpkg"),
        dtype={"geo_value": str})
    pd.testing.assert_frame_equal(expected1, output1)

    # test left join
    output2 = plotting._join_msa_geo_df(test_input, "msa", geo_info, "left")
    expected2 = gpd.read_file(
        os.path.join(CURRENT_PATH, "../reference_data/expected__join_msa_geo_df_left.gpkg"),
        dtype={"geo_value": str})
    pd.testing.assert_frame_equal(expected2, output2)


def test__join_hrr_geo_df():
    test_input = pd.DataFrame({"hrr": ["1", "102", "96"],
                               "test_value": [1.5, 2.5, 3],
                               "test_value2": [21.5, 32.5, 34]})
    geo_info = gpd.read_file(os.path.join(CURRENT_PATH, SHAPEFILE_PATHS["hrr"]))
    # test right join
    output1 = plotting._join_hrr_geo_df(test_input, "hrr", geo_info)
    assert type(output1) is gpd.GeoDataFrame

    expected1 = gpd.read_file(
        os.path.join(CURRENT_PATH, "../reference_data/expected__join_hrr_geo_df_right.gpkg"),
        dtype={"geo_value": str})
    pd.testing.assert_frame_equal(expected1, output1)

    # test left join
    output2 = plotting._join_hrr_geo_df(test_input, "hrr", geo_info, "left")
    expected2 = gpd.read_file(
        os.path.join(CURRENT_PATH, "../reference_data/expected__join_hrr_geo_df_left.gpkg"),
        dtype={"geo_value": str})
    pd.testing.assert_frame_equal(expected2, output2)



import pytest

from covidcast import geography


@pytest.mark.parametrize("test_key, test_kwargs, expected", [
    (
            "not a fips",
            {},
            [None]
    ),
    (
            "42003",
            {},
            ["Allegheny County"],
    ),
    (
            "4200",
            {"ties_method": "all"},
            [{"42000": ["Pennsylvania"], "42001": ["Adams County"],
              "42003": ["Allegheny County"], "42005": ["Armstrong County"],
              "42007": ["Beaver County"], "42009": ["Bedford County"]}]
    ),
    (
            "4200",
            {},
            ["Pennsylvania"]
    )
])
def test_fips_to_name(test_key, test_kwargs, expected):
    assert geography.fips_to_name(test_key, **test_kwargs) == expected


@pytest.mark.parametrize("test_key, test_kwargs, expected", [
    (
            "not a cbsa",
            {},
            [None]
    ),
    (
            "38300",
            {},
            ["Pittsburgh, PA"]
    ),
    (
            "389",
            {"ties_method": "all"},
            [{"38900": ["Portland-Vancouver-Hillsboro, OR-WA"], "38940": ["Port St. Lucie, FL"]}]
    ),
    (
            ["38300", "389"],
            {"ties_method": "all"},
            [
                {"38300": ["Pittsburgh, PA"]},
                {"38900": ["Portland-Vancouver-Hillsboro, OR-WA"], "38940": ["Port St. Lucie, FL"]}
            ]

    )
])
def test_cbsa_to_name(test_key, test_kwargs, expected):
    assert geography.cbsa_to_name(test_key, **test_kwargs) == expected


@pytest.mark.parametrize("test_key, test_kwargs, expected", [
    (
            "CA",
            {},
            ["California"]
    ),
    (
            "CAA",
            {},
            [None]
    ),
    (
            ["CA", "PA"],
            {},
            ["California", "Pennsylvania"]
    ),
    (
            ["CAAA", "PA"],
            {},
            [None, "Pennsylvania"]
    ),
])
def test_abbr_to_name(test_key, test_kwargs, expected):
    assert geography.abbr_to_name(test_key, **test_kwargs) == expected


@pytest.mark.parametrize("test_key, test_kwargs, expected", [
    (
            "California",
            {},
            ["CA"]
    ),
    (
            "Californiaaaaa",
            {},
            [None]
    ),
    (
            ["California", "Pennsylvania"],
            {},
            ["CA", "PA"]
    ),
    (
            ["California", "Pennsylvaniaa"],
            {},
            ["CA", None]
    ),
    (
            ["California", "Pennsylvania", "California"],
            {},
            ["CA", "PA", "CA"]
    ),
])
def test_name_to_abbr(test_key, test_kwargs, expected):
    assert geography.name_to_abbr(test_key, **test_kwargs) == expected


@pytest.mark.parametrize("test_key, test_kwargs, expected", [
    (
            "Pittsburgh",
            {},
            ["38300"]
    ),
    (
            "New",
            {"state": "CA"},
            [None]
    ),
    (
            "New",
            {},
            ["14460"]
    )
])
def test_name_to_cbsa(test_key, test_kwargs, expected):
    assert geography.name_to_cbsa(test_key, **test_kwargs) == expected


@pytest.mark.parametrize("test_key, test_kwargs, expected", [
    (
            "Allegheny",
            {},
            ["42003"]
    ),
    (
            "Miami",
            {},
            ["12086"]
    ),
    (
            "Miami",
            {"ties_method": "all"},
            [{"Miami-Dade County": ["12086"], "Miami County": ["18103", "20121", "39109"]}]
     ),
    (
            ["Allegheny", "Miami", "New "],
            {"ties_method": "all"},
            [
                {"Allegheny County": ["42003"]},
                {"Miami-Dade County": ["12086"],
                 "Miami County": ["18103", "20121", "39109"]},
                {"New Haven County": ["09009"], "New London County": ["09011"],
                 "New Castle County": ["10003"], "New Madrid County": ["29143"],
                 "New York County": ["36061"], "New Hanover County": ["37129"],
                 "New Kent County": ["51127"], "New Hampshire": ["33000"],
                 "New Jersey": ["34000"], "New Mexico": ["35000"],
                 "New York": ["36000"]}
            ]
    ),
    (
            ["Allegheny", "Miami", "New "],
            {},
            ["42003", "12086", "09009"]
    ),
    (
            "New ",
            {"ties_method": "all", "state": "ny"},
            [{"New York": ["36000"], "New York County": ["36061"]}]
    ),
])
def test_name_to_fips(test_key, test_kwargs, expected):
    assert geography.name_to_fips(test_key, **test_kwargs) == expected


@pytest.mark.parametrize("test_args, test_kwargs, expected", [
    (
            (["a", "b"], ["a", "b", "c"], ["x", "y", "z"]),
            {},
            ["x", "y"]
    ),
    (
            (["a", "b"], ["a", "a", "aa", "b"], ["w", "x", "y", "z"]),
            {"ties_method": "all"},
            [{"a": ["w", "x"], "aa": ["y"]}, {"b": ["z"]}]
    ),
    (
            (["a", "b"], ["A", "aa", "b"], ["x", "y", "z"]),
            {"ignore_case": True},
            ["x", "z"]
    ),
    (
            (["a", "b"], ["a", "aa", "b"], ["x", "y", "z"]),
            {"fixed": True},
            ["x", "z"]
    ),
    (
            (["a", "b"], ["a", "aa", "b"], ["x", "y", "z"]),
            {"ties_method": "all", "fixed": True},
            [{"a": ["x"]}, {"b": ["z"]}]
    ),
    (
            (["a", "b"], ["A", "aa", "b"], ["x", "y", "z"]),
            {"ties_method": "all", "ignore_case": True},
            [{"A": ["x"], "aa": ["y"]}, {"b": ["z"]}]
    ),
    (
            (["a", "b"], ["A", "aa", "b"], ["x", "y", "z"]),
            {"ties_method": "all", "fixed": True, "ignore_case": True},
            [{}, {"b": ["z"]}]
    ),
    (
            ("A", ["aa", "a"], ["x", "y"]),
            {},
            [None]
    ),
    (
            ("A", ["aa", "a"], ["x", "y"]),
            {"ignore_case": True},
            ["x"]
    ),
    (
            (["A"], ["aa", "a"], ["x", "y"]),
            {"ignore_case": True},
            ["x"]
    ),
    (
            ("A", ["aa", "a"], ["x", "y"]),
            {"ignore_case": True, "ties_method": "all"},
            [{"aa": ["x"], "a": ["y"]}]
    ),
    (
            (["A"], ["aa", "a"], ["x", "y"]),
            {"ignore_case": True, "ties_method": "all"},
            [{"aa": ["x"], "a": ["y"]}]
    ),
    (
            ("A", ["aa", "a"], ["x", "y"]),
            {"ignore_case": True, "ties_method": "all"},
            [{"aa": ["x"], "a": ["y"]}]
    ),
    (
            ("A", ["aa", "a"], ["x", "y"]),
            {"ignore_case": True},
            ["x"]
    ),
    (
            ("a", ["a", "a"], ["x", "y"]),
            {"ties_method": "all"},
            [{"a": ["x", "y"]}]
    ),
    (
            ("a", ["aa", "a"], ["x", "y"]),
            {"ties_method": "all"},
            [{"aa": ["x"], "a": ["y"]}]
    ),
    (
            ("a", ["aa", "a"], ["x", "y"]),
            {"ties_method": "all", "fixed": True},
            [{"a": ["y"]}]
    ),
    (
            ("A", ["aa", "a"], ["x", "y"]),
            {"ties_method": "all", "fixed": True, "ignore_case": True},
            [{}]
    ),
    (
            ("A", ["a", "a"], ["x", "y"]),
            {"ties_method": "all", "ignore_case": True},
            [{"a": ["x", "y"]}]
    )
])
def test__lookup(test_args, test_kwargs, expected):
    assert geography._lookup(*test_args, **test_kwargs) == expected
    with pytest.raises(ValueError):
        geography._lookup(None, None, None, ties_method="not a real method")


@pytest.mark.parametrize("test_dict_list, expected_return, warn, expected_warning", [
    (
            [{"a": ["x", "y"]}],
            ["x"],
            1,
            "Some inputs were not uniquely matched; returning only the first match in "
            "each case. To return all matches, set `ties_method='all'`"
    ),
    (
            [{"a": ["x", "y"], "b": ["i", "j", "k"]}],
            ["x"],
            1,
            "Some inputs were not uniquely matched; returning only the first match in "
            "each case. To return all matches, set `ties_method='all'`"
    ),
    (
            [{"a": ["x", "y"]}, {"b": ["i", "j", "k"]}],
            ["x", "i"],
            1,
            "Some inputs were not uniquely matched; returning only the first match in "
            "each case. To return all matches, set `ties_method='all'`"
    ),
    (
            [{"a": ["x"]}],
            ["x"],
            0,
            None
    )
])
def test__get_first_tie(test_dict_list, expected_return, warn, expected_warning):
    if warn:
        with pytest.warns(UserWarning) as record:
            assert geography._get_first_tie(test_dict_list) == expected_return
            assert record[0].message.args[0] == expected_warning
    else:
        assert geography._get_first_tie(test_dict_list) == expected_return




"""Fetch data from Delphi's COVIDcast API.

The COVIDcast API provides daily updated data on the COVID-19 pandemic in the
United States, including cases, deaths, medical records, nationwide symptom
surveys, and other data collated by the Delphi research group at Carnegie Mellon
University.

Functions:

* signal - Fetch a Pandas data frame for one signal.
* metadata - Fetch metadata for all available signals.

"""

from .covidcast import signal, metadata, aggregate_signals
from .plotting import plot_choropleth, get_geo_df, animate
from .geography import (fips_to_name, cbsa_to_name, abbr_to_name,
                        name_to_abbr, name_to_cbsa, name_to_fips)



from datetime import date, datetime
from unittest.mock import patch

# Force tests to use a specific backend, so they reproduce across platforms
import matplotlib
matplotlib.use("AGG")

import pandas as pd
import numpy as np
import pytest
from covidcast import covidcast


def sort_df(df):
    """Helper function for sorting dfs for comparison."""
    df = df.sort_index(axis=1)
    df.sort_values(df.columns[0], inplace=True)
    return df


@patch("covidcast.covidcast._signal_metadata")
@patch("delphi_epidata.Epidata.covidcast")
def test_signal(mock_covidcast, mock_metadata):
    mock_covidcast.return_value = {"result": 1,  # successful API response
                                   "epidata": [{"time_value": 20200622, "issue": 20200724}],
                                   "message": "success"}
    mock_metadata.return_value = {"max_time": pd.Timestamp("2020-08-04 00:00:00"),
                                  "min_time": pd.Timestamp("2020-08-03 00:00:00")}

    return_rows = {"time_value": [datetime(2020, 6, 22)],
                   "issue": datetime(2020, 7, 24),
                   "geo_type": "county",
                   "data_source": "source",
                   "signal": "signal"
                   }
    # test happy path with no start or end day and one geo_value
    response = covidcast.signal("source", "signal", geo_values="CA")
    expected = pd.DataFrame(return_rows, index=[0]*2)
    assert sort_df(response).equals(sort_df(expected))

    # test happy path with no start or end day and two geo_values
    response = covidcast.signal("source", "signal", geo_values=["CA", "AL"])
    expected = pd.DataFrame(return_rows, index=[0]*4)
    assert sort_df(response).equals(sort_df(expected))

    # test happy path with start and end day (8 days apart) and one geo_value
    response = covidcast.signal("source", "signal", start_day=date(2020, 8, 1),
                                end_day=date(2020, 8, 8), geo_values="CA")
    expected = pd.DataFrame(return_rows, index=[0]*8)
    assert sort_df(response).equals(sort_df(expected))

    # test duplicate geo values
    response = covidcast.signal("source", "signal", start_day=date(2020, 8, 1),
                                end_day=date(2020, 8, 8), geo_values=["CA", "CA"])
    expected = pd.DataFrame(return_rows, index=[0]*8)

    assert sort_df(response).equals(sort_df(expected))

    # test no df output
    assert not covidcast.signal("source", "signal", geo_values=[])

    # test incorrect geo
    with pytest.raises(ValueError):
        covidcast.signal("source", "signal", geo_type="not_a_real_geo")

    # test invalid dates
    with pytest.raises(ValueError):
        covidcast.signal("source", "signal", geo_type="state",
                         start_day=date(2020, 4, 2), end_day=date(2020, 4, 1))


@patch("delphi_epidata.Epidata.covidcast_meta")
def test_metadata(mock_covidcast_meta):
    # not generating full DF since most attributes used
    mock_covidcast_meta.side_effect = [{"result": 1,  # successful API response
                                        "epidata": [{"max_time": 20200622, "min_time": 20200421},
                                                    {"max_time": 20200724, "min_time": 20200512}],
                                        "message": "success"},
                                       {"result": 0,  # unsuccessful API response
                                        "epidata": [{"max_time": 20200622, "min_time": 20200421},
                                                    {"max_time": 20200724, "min_time": 20200512}],
                                        "message": "error: failed"}]

    # test happy path
    response = covidcast.metadata()
    expected = pd.DataFrame({"max_time": [datetime(2020, 6, 22), datetime(2020, 7, 24)],
                                     "min_time": [datetime(2020, 4, 21), datetime(2020, 5, 12)]})
    assert sort_df(response).equals(sort_df(expected))

    # test failed response raises RuntimeError
    with pytest.raises(RuntimeError):
        covidcast.metadata()


def test_aggregate_signals():
    test_input1 = pd.DataFrame(
        {"geo_value": ["a", "b", "c", "a"],
         "time_value": [date(2020, 1, 1), date(2020, 1, 1), date(2020, 1, 1), date(2020, 1, 2)],
         "value": [2, 4, 6, 8],
         "signal": ["i", "i", "i", "i"],
         "geo_type": ["state", "state", "state", "state"],
         "data_source": ["x", "x", "x", "x"]})
    test_input2 = pd.DataFrame(
        {"geo_value": ["a", "b", "c", "d"],
         "time_value": [date(2020, 1, 1), date(2020, 1, 1), date(2020, 1, 1), date(2020, 1, 1)],
         "value": [1, 3, 5, 7],
         "signal": ["j", "j", "j", "j"],
         "geo_type": ["state", "state", "state", "state"],
         "data_source": ["y", "y", "y", "y"],
         "extra_col": ["0", "0", "0", "0"]})
    test_input3 = pd.DataFrame(
        {"geo_value": ["b", "c", "d", "b"],
         "time_value": [date(2020, 1, 1), date(2020, 1, 1), date(2020, 1, 1), date(2020, 1, 2)],
         "value": [0.5, 1.5, 2.5, 3.5],
         "signal": ["k", "k", "k", "k"],
         "geo_type": ["state", "state", "state", "state"],
         "data_source": ["z", "z", "z", "z"]})
    # test 3 signals from 3 sources with outer join
    expected1 = pd.DataFrame(
        {"geo_value": ["a", "b", "c", "d", "a", "b", "c", "d", "b"],
         "time_value": [date(2020, 1, 1), date(2020, 1, 1), date(2020, 1, 1), date(2020, 1, 1),
                        date(2020, 1, 2), date(2020, 1, 2), date(2020, 1, 2), date(2020, 1, 2),
                        date(2020, 1, 3)],
         "x_i_0_value": [2, 4, 6, np.nan, 8, np.nan, np.nan, np.nan, np.nan],
         "y_j_1_value": [1, 3, 5, 7, np.nan, np.nan, np.nan, np.nan, np.nan],
         "y_j_1_extra_col": ["0", "0", "0", "0", np.nan, np.nan, np.nan, np.nan, np.nan],
         "z_k_2_value": [np.nan, np.nan, np.nan, np.nan, np.nan, 0.5, 1.5, 2.5, 3.5],
         "geo_type": ["state"]*9})
    assert covidcast.aggregate_signals(
        [test_input1, test_input2, test_input3], dt=[0, 0, 1]).equals(expected1)

    # test 3 signals from 3 sources with inner join has no intersection
    assert covidcast.aggregate_signals(
        [test_input1, test_input3], dt=[0, 1], join_type="inner").empty

    # test 2 signals from same source (one lagged) with inner join
    expected2 = pd.DataFrame(
        {"geo_value": ["a"],
         "time_value": [date(2020, 1, 2)],
         "x_i_0_value": [8],
         "x_i_1_value": [2],
         "geo_type": ["state"]})
    assert covidcast.aggregate_signals(
        [test_input1, test_input1], dt=[0, 1], join_type="inner").equals(expected2)

    # test same signal twice with a lag
    expected3 = pd.DataFrame(
        {"geo_value": ["a", "b", "c", "a", "b", "c", "a"],
         "time_value": [date(2020, 1, 1), date(2020, 1, 1), date(2020, 1, 1), date(2020, 1, 2),
                        date(2020, 1, 2), date(2020, 1, 2), date(2020, 1, 3)],
         "x_i_0_value": [2, 4, 6, 8, np.nan, np.nan, np.nan],
         "x_i_1_value": [np.nan, np.nan, np.nan, 2, 4, 6, 8],
         "geo_type": ["state"]*7})

    assert covidcast.aggregate_signals([test_input1, test_input1], dt=[0, 1]).equals(expected3)

    # test invalid lag length
    with pytest.raises(ValueError):
        covidcast.aggregate_signals([test_input1, test_input1], dt=[0])

    # test mixed geo_types
    test_input4 = pd.DataFrame(
        {"geo_value": ["b", "c", "d", "b"],
         "time_value": [date(2020, 1, 1), date(2020, 1, 1), date(2020, 1, 1), date(2020, 1, 2)],
         "value": [0.5, 1.5, 2.5, 3.5],
         "signal": ["k", "k", "k", "k"],
         "geo_type": ["county", "county", "county", "county"],
         "data_source": ["z", "z", "z", "z"]})
    with pytest.raises(ValueError):
        covidcast.aggregate_signals([test_input1, test_input4])


def test__detect_metadata():
    test_input = pd.DataFrame(
        {"data_source": ["a", "a"], "signal": ["b", "b"], "geo_type": ["c", "c"]})
    assert covidcast._detect_metadata(test_input) == ("a", "b", "c")

    # test heterogenous cases error
    test_bad_source = pd.DataFrame(
        {"data_source": ["a", "d"], "signal": ["b", "b"], "geo_type": ["c", "c"]})
    with pytest.raises(ValueError):
        covidcast._detect_metadata(test_bad_source)
    test_bad_signal = pd.DataFrame(
        {"data_source": ["a", "a"], "signal": ["d", "b"], "geo_type": ["c", "c"]})
    with pytest.raises(ValueError):
        covidcast._detect_metadata(test_bad_signal)
    test_bad_geo = pd.DataFrame(
        {"data_source": ["a", "a"], "signal": ["b", "b"], "geo_type": ["c", "x"]})
    with pytest.raises(ValueError):
        covidcast._detect_metadata(test_bad_geo)


@patch("delphi_epidata.Epidata.covidcast")
def test__fetch_single_geo(mock_covidcast):
    # not generating full DF since most attributes used
    mock_covidcast.side_effect = [{"result": 1,  # successful API response
                                   "epidata": [{"time_value": 20200622, "issue": 20200724}],
                                   "message": "success"},
                                  {"result": 1,  # second successful API
                                   "epidata": [{"time_value": 20200821, "issue": 20200925}],
                                   "message": "success"},
                                  {"message": "error: failed"},  # unsuccessful API response
                                  {"message": "success"}]  # no epidata

    # test happy path with 2 day range
    response = covidcast._fetch_single_geo(
        None, None, date(2020, 4, 2), date(2020, 4, 3), None, None, None, None, None)
    expected = pd.DataFrame({"time_value": [datetime(2020, 6, 22), datetime(2020, 8, 21)],
                             "issue": [datetime(2020, 7, 24), datetime(2020, 9, 25)],
                             "geo_type": None,
                             "data_source": None,
                             "signal": None
                             },
                            index=[0, 0])
    assert sort_df(response).equals(sort_df(expected))

    # test warning is raised if unsuccessful API response
    with pytest.warns(UserWarning):
        covidcast._fetch_single_geo(None, None, date(2020, 4, 2), date(2020, 4, 2),
                                    None, None, None, None, None)

    # test no epidata yields nothing
    assert not covidcast._fetch_single_geo(None, None, date(2020, 4, 2), date(2020, 4, 1),
                                           None, None, None, None, None)

    # test end_day < start_day yields nothing
    assert not covidcast._fetch_single_geo(None, None, date(2020, 4, 2), date(2020, 4, 1),
                                           None, None, None, None, None)


@patch("covidcast.covidcast.metadata")
def test__signal_metadata(mock_metadata):
    mock_metadata.return_value = pd.DataFrame({"data_source": ["usa-facts", "doctor-visits"],
                                               "signal": ["raw_cli", "smooth_cli"],
                                               "time_type": ["day", "day"],
                                               "geo_type": ["hrr", "state"]})

    # test happy path
    assert covidcast._signal_metadata("usa-facts", "raw_cli", "hrr") == \
           {"data_source": "usa-facts", "signal": "raw_cli", "time_type": "day", "geo_type": "hrr"}

    # test no matches raises ValueError
    with pytest.raises(ValueError):
        covidcast._signal_metadata("not", "actual", "values")


def test__date_to_api_string():
    # since the function just wraps strftime, this is just to ensure the format doesn't change
    assert covidcast._date_to_api_string(date(2020, 4, 2)) == "20200402"


def test__dates_to_api_strings():
    # test happy paths
    assert covidcast._dates_to_api_strings(date(2020, 4, 2)) == "20200402"
    assert covidcast._dates_to_api_strings([date(2020, 4, 2),
                                            date(2020, 5, 2)]) == "20200402-20200502"
